
package br.ufsc.das.paxosatwar.messages;

import br.ufsc.das.util.CommunicationSystem;
import br.ufsc.das.util.SystemMessage;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.Arrays;

/**
 * This class represents a message used in the paxos protocol.
 */

public class PaxosMessage extends SystemMessage {
    
    private int number;
    private int round;
    private int from;
    private Object value;
    private Object proof;
    
    private int paxosType;
    
    public PaxosMessage(){}
    
    public PaxosMessage(int paxosType, int id,int round,int from, Object value, Object proof){
        super(CommunicationSystem.PAXOS_MSG);
        this.paxosType = paxosType;
        this.number = id;
        this.round = round;
        this.from = from;
        this.value = value;
        this.proof = proof;
    }
    
    public PaxosMessage(int paxosType, int id,int round,int from, Object value){
        this(paxosType, id, round, from, value, null);
    }
    
    public PaxosMessage(int paxosType, int id,int round, int from){
        this(paxosType, id, round, from, null, null);
    }

    
    public void writeExternal(ObjectOutput out)
    throws IOException{
        super.writeExternal(out);
        out.writeInt(number);
        out.writeInt(round);
        out.writeInt(from);
        out.writeInt(paxosType);
        out.writeObject(value);
        out.writeObject(proof);
    }
    
    public void readExternal(ObjectInput in)
    throws IOException,
            ClassNotFoundException{
        super.readExternal(in);
        number = in.readInt();
        round = in.readInt();
        from = in.readInt();
        paxosType = in.readInt();
        value = in.readObject();
        proof = in.readObject();
    }
    
    
    
    public int getRound() {
        return round;
    }
    
    public Object getValue() {
        return value;
    }
    
    public Object getProof() {
        return proof;
    }
    
    public int getNumber() {
        return number;
    }
    
    public int getFrom() {
        return from;
    }
    
    public int getPaxosType() {
        return paxosType;
    }
    
    public String getPaxosVerboseType() {
        if (paxosType==MessageFactory.COLLECT)
            return "COLLECT";
        else if (paxosType==MessageFactory.DECIDE)
            return "DECIDE";
        else if (paxosType==MessageFactory.FREEZE)
            return "FREEZE";
        else if (paxosType==MessageFactory.PROPOSE)
            return "PROPOSE";
        else if (paxosType==MessageFactory.STRONG)
            return "STRONG";
        else if (paxosType==MessageFactory.WEAK)
            return "WEAK";
        else
            return "";
    }
    
    public String toString() {
        return "type="+getPaxosVerboseType()+", number="+getNumber()+", round="+getRound()+", from="+getFrom()+", "+
                ((getProof() instanceof CollectProof[])?
                    Arrays.asList((CollectProof[])getProof()):getProof());
    }
}
