

package br.ufsc.das.paxosatwar.roles.util;

import br.ufsc.das.util.Configuration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import br.ufsc.das.paxosatwar.Consensus;
import java.util.concurrent.locks.ReentrantLock;



public class Execution {
    
    private ExecutionManager manager;
    private int n;
    private int eid;
    private int[] proposers;
    private int[] otherProcesses;
    private Consensus learner;
    private Map<Integer,Round> rounds = new HashMap<Integer,Round>();
    private boolean decided;
    private long initialTimeout;
    private int decidedRound = -1;
    private ReentrantLock lock;
    public static int lastExec = 0;
    public static long numDecides = 0;
    public static long firstDecideTimeInstant;    
    
    /** Creates a new instance of Execution for Acceptor Manager */
    protected Execution(ExecutionManager manager, int eid, Consensus learner, long initialTimeout) {
        if (eid!=0 && lastExec==eid){
            //br.ufsc.das.util.Logger.println("Vou criar Execution apesar de jah existir um com o mesmo eid="+eid);
        }
        lastExec = eid;
        this.manager = manager;
        this.eid = eid;
        this.learner = learner;
        this.initialTimeout = initialTimeout;
        this.lock = new ReentrantLock();
        //br.ufsc.das.util.Logger.println("A criar Execution, eid="+eid);
    }
    
    public int getId() {
        return eid;
    }

    
    public ExecutionManager getManager() {
        return manager;
    }

    public Consensus getLearner() {
        return learner;
    }
    
    public Round getRound(int number) {
        lock.lock();
        Round round = rounds.get(number);
        while(round == null) {
            round = new Round(this,number, initialTimeout);
            rounds.put(number,round);
        }
        lock.unlock();
        return round;
    }
    
    public void removeRounds(int limit){
        lock.lock();
        int num = rounds.size();
        for(int i = limit+1; i < num; i++){
            this.rounds.remove(i).setRemoved();
        }
        lock.unlock();
    }
    
    public Round getDecidedRound(){
        lock.lock();
        Round r = rounds.get(decidedRound);
        lock.unlock();
        return r; 
    }
    
    public Round getLastRound(){
        lock.lock();
        Round r = rounds.get(rounds.size()-1); 
        lock.unlock();
        return r;
    }
    
    public boolean isDecided() {
        return decided;
    }

    public void decide(Object value, int round) {
        lock.lock();              
        
        if(!decided){
            //br.ufsc.das.util.Logger.println("(Execution) decide!, round="+round+" execution="+eid);
            decided = true;            
            decidedRound = round;
            lock.unlock();
            if(learner != null){
                learner.learn(value);
            }
            else {               
                //br.ufsc.das.util.Logger.println("(Execution) Problema: o learner ´eh null!");
            }
        }
        else
            lock.unlock();
    }
}

