

package br.ufsc.das.paxosatwar.roles.util;

import br.ufsc.das.paxosatwar.messages.MessageFactory;
import br.ufsc.das.tom.usage.TOMLayer;
import br.ufsc.das.util.Configuration;
import br.ufsc.das.util.TOMConfiguration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.Semaphore;
import br.ufsc.das.paxosatwar.roles.Acceptor;
import br.ufsc.das.paxosatwar.roles.Proposer;
import br.ufsc.das.paxosatwar.messages.PaxosMessage;
import br.ufsc.das.paxosatwar.Consensus;
import java.util.Collections;
import java.util.concurrent.locks.ReentrantLock;

public class ExecutionManager {
    
    private Acceptor acceptor;
    private Proposer proposer;
    private int me;
    private int[] acceptors;
    private int[] otherAcceptors;
    private Map<Integer,Execution> executions = new TreeMap<Integer,Execution>();
    private Map<Integer,List<PaxosMessage>> outOfContext = new HashMap<Integer,List<PaxosMessage>>();
    private Map<Integer,PaxosMessage> pendentProp = new HashMap<Integer,PaxosMessage>();
    
    private boolean stoped = false;
    private List<PaxosMessage> stopedMsgs = new LinkedList<PaxosMessage>();
    
    public final int quorumF;
    public final int quorum2F;
    public final int quorumStrong;
    public final int quorumFastDecide;
    private TOMLayer multicastLayer;
    private long initialTimeout;

    private ReentrantLock lock1;
    //private ReentrantLock lock2;
    
    private Round stopedRound = null;
    /**
     * Creates a new instance of ExecutionManager
     */
    public ExecutionManager(Acceptor acceptor, Proposer proposer,
            int[] acceptors, int f, int me, long initialTimeout) {
        this.acceptor = acceptor;
        this.proposer = proposer;
        this.acceptors = acceptors;
        this.me = me;
        this.initialTimeout = initialTimeout;
        this.quorumF = f;
        this.quorum2F = 2*f;
        this.quorumStrong = (int)Math.ceil((acceptors.length+f)/2);
        this.quorumFastDecide = (int)Math.ceil((acceptors.length+3*f)/2);
        lock1 = new ReentrantLock();
        br.ufsc.das.util.Logger.startInstant = System.currentTimeMillis();
        //br.ufsc.das.util.Logger.println("(ExecutionManager) Criei ExecutionManager");
    }
    
    public void setTOMLayer(TOMLayer tom){
        this.multicastLayer = tom;
    }
    
    public void stop(){
       this.stoped = true;
       if(multicastLayer.getInExec() != -1){
            stopedRound = getExecution(multicastLayer.getInExec()).getLastRound();
            stopedRound.getTimeoutTask().cancel();
       }
    }
    
    public void restar(){
        this.stoped = false;
        if(stopedRound != null){
            acceptor.scheduleTimeout(stopedRound);
            stopedRound = null;
        }
        acceptor.getMEZone();
        for(int i = 0; i < stopedMsgs.size();i++){
            PaxosMessage msg = stopedMsgs.get(i);
            if(msg.getPaxosType() == MessageFactory.PROPOSE){
                acceptor.proposeReceived(msg);
            }else{
                acceptor.processMessage(msg,getExecution(msg.getNumber()));
            }
        }
        acceptor.leaveMEZone();
    }
    
    public boolean checkLimits(PaxosMessage msg, TOMConfiguration conf){
        int consId = msg.getNumber();
        int last = multicastLayer.getLastExec();
        int low = 0;
        if(multicastLayer.lastDecided != null){
            low = multicastLayer.lastDecided.getId();
        }
        
        lock1.lock();
        if(low <= consId  &&   ((last + conf.getPaxosHighMarc()) > consId)){
            //br.ufsc.das.util.Logger.println("(checkLimits) ("+consId+") Mensagem recebida estah dentro dos limites");    
            if(stoped){
                //br.ufsc.das.util.Logger.println("(checkLimits) ("+consId+") Vou adicionar ao stopped");  
                stopedMsgs.add(msg); 
                lock1.unlock();
                return false;               
            }else if(consId > (last+1)){                
                if(msg.getPaxosType() == MessageFactory.PROPOSE){
                    //br.ufsc.das.util.Logger.println("(checkLimits) ("+consId+") eh um out of context propose, should be " + (last+1));                    
                    pendentProp.put(consId,msg);                    
                }else{
                    //br.ufsc.das.util.Logger.println("(checkLimits) ("+consId+") eh uma out of context msg (nao propose) de "+msg.getFrom()+", should be " + (last+1));
                    addOutOfContextMessage(msg);
                }
                lock1.unlock();
                return false;
            }else{
                //br.ufsc.das.util.Logger.println("(checkLimits) ("+consId+") msg dentro do contexto");
                lock1.unlock();
                return true;
            }
        }
        lock1.unlock();
        //br.ufsc.das.util.Logger.println("(checkLimits) Mensagem recebida nao estah dentro dos limites");    
        return false;
    }
    
    public int getProcessId() {
        return me;
    }
    
    public int[] getAcceptors() {
        return acceptors;
    }
    
    public int[] getOtherAcceptors() {
        if(otherAcceptors == null) {
            otherAcceptors = new int[acceptors.length-1];
            int c = 0;
            for(int i=0; i<acceptors.length; i++) {
                if(acceptors[i] != me) {
                    otherAcceptors[c++] = acceptors[i];
                }
            }
        }
        return otherAcceptors;
    }
    
    public Acceptor getAcceptor() {
        return acceptor;
    }
    
    public Proposer getProposer() {
        return proposer;
    }
    
    public void checkNextExecution(int eid){
        lock1.lock();
        if(pendentProp.get(eid) != null || outOfContext.get(eid) != null){
            //o unlock eh feito dentro da funcao
            createExecution(eid);
        }
        else
            lock1.unlock();
    }
    
    /** must be called only by Acceptor */
    private Execution createExecution(int eid) {
        Execution execution = executions.get(eid);
        //lock1.lock();
        if(execution != null) {
            //execution = (Execution)executions.get(eid);
            lock1.unlock();
            return execution;
            //throw new IllegalArgumentException("eid="+eid+" already in use");
        }
        Consensus learner = new Consensus(multicastLayer,proposer, eid);
        execution = new Execution(this,eid,learner,initialTimeout);
        executions.put(eid, execution);
        PaxosMessage prop = pendentProp.remove(eid);
        if(prop != null){            
            acceptor.proposeReceived(prop);            
        }
        List<PaxosMessage> messages = outOfContext.remove(eid);
        lock1.unlock();
        if(messages != null) {
            //br.ufsc.das.util.Logger.println("(createExecution) ("+eid+") A processar " + messages.size() + " mensagens recebidas previamente fora de contexto");
            for(Iterator<PaxosMessage> i = messages.iterator(); i.hasNext(); ) {
                acceptor.processMessage(i.next(), execution);                
            }
            //br.ufsc.das.util.Logger.println("(createExecution) ("+eid+") Terminei processamento de mensagens recebidas previamente fora de contexto");
        }
        else {
            //br.ufsc.das.util.Logger.println("(createExecution) ("+eid+") Nao ha msgs fora de contexto");
        }
        //lock1.unlock();
        return execution;
    }
    
    public void setStable(int id){
        lock1.lock();
        executions.remove(id);
        pendentProp.remove(id);
        outOfContext.remove(id);
        lock1.unlock();
    }
    
    public Execution removeExecution(int id){
        lock1.lock();
        pendentProp.remove(id);
        outOfContext.remove(id);
        Execution execution = executions.remove(id);
        lock1.unlock();
        return execution;
    }
    
    
    public Execution getExecution(int eid) {
        lock1.lock();
        //o unlock eh feito dentro da funcao
        Execution ret = createExecution(eid);               
        //lock1.unlock();
        return ret;
    }
    
    public Execution isInExecution(int eid) {
        lock1.lock();
        Execution execution = executions.get(eid);
        lock1.unlock();
        return execution;
    }
    
    private void addOutOfContextMessage(PaxosMessage m) {
        
        List<PaxosMessage> messages = outOfContext.get(m.getNumber());
        if(messages == null) {
            messages = new LinkedList<PaxosMessage>();
            outOfContext.put(m.getNumber(), messages);
        }
        messages.add(m);
       
    }
}
