/*
 * TOMClientPerformanceTest.java
 *
 * Created on 22 de Maio de 2006, 14:57
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package br.ufsc.das.tom.demo;

import br.ufsc.das.tom.usage.TOMReplyReceiver;
import br.ufsc.das.tom.TOMSender;
import br.ufsc.das.tom.util.TOMMessage;
import br.ufsc.das.util.Storage;
import br.ufsc.das.util.TOMConfiguration;
import br.ufsc.das.util.CommunicationSystem;

import java.util.concurrent.Semaphore;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author edualchieri
 */
public class TOMClientLatency extends TOMSender{
    
    private Storage st;
    private int exec;
    private int argSize;
    private Semaphore sm = new Semaphore(0);
    private Semaphore mutex = new Semaphore(1);
    private int id = 0;
    private int count = 0;
    private CommunicationSystem cs;
    private int f;
    private int n;
    private int currentId = 0;
    private long last_send_instant = 0;
    private int num_sends = 0;
    private int interval = 0;
        
    public TOMClientLatency(int exec, int argSize, int interval) {
        this.exec = exec;
        this.argSize = argSize;
        this.st = new Storage(exec/2);
        //id for this client
        this.id = (int) System.currentTimeMillis();
        this.currentId = id;
        //create the configuration object
        TOMConfiguration conf = new TOMConfiguration(id);
        //create the communication system
        cs = new CommunicationSystem(conf);
        //the number of suported faults
        this.f = conf.getF();
        this.n = conf.getN();
        this.interval = interval;
         br.ufsc.das.util.Logger.startInstant = System.currentTimeMillis();
        //initialize the TOM sender
        this.init(cs, conf);
    }
    
    private void run(){
        long totalBegin = System.nanoTime();
        
        for(int i = 0; i < exec; i++){
            try {
                int currId = currentId;
                num_sends = i;
                Object[] obj = new Object[2];
                obj[0] = currId;
                obj[1] = new byte[argSize];
                //obj[0] = id;
                //obj[1] = "Eduardo"+id;
                //obj[1] = "eu sou cliente";
                //obj[2] = "confidencialidade";
                if (i%1000==0 && i!=0)
                    System.out.println("Enviando " + i + " / "+ exec);
                //br.ufsc.das.util.Logger.println(", currentId = " + currentId);
                last_send_instant = System.nanoTime();
                this.TOMulticast(obj);
                
                try {
                    this.sm.acquire();
                } catch (Exception e) {
                    e.printStackTrace();
                }
                            

                if (interval > 0){
                    //sleeps interval ms before sending next request
                    Thread.sleep(interval);
                }

                //------------------------------------------------//
            } catch (InterruptedException ex) {
                Logger.getLogger(TOMClientLatency.class.getName()).log(Level.SEVERE, null, ex);
            }
           
            //------------------------------------------------//
        }
        long totalElapsedTime = System.nanoTime()-totalBegin;
        System.out.println("Average time for "+exec/2+" executions (-10%) = "+this.st.getAverage(true)/1000+ " us ");
        System.out.println("Standard desviation for "+exec/2+" executions (-10%) = "+this.st.getDP(true)/1000 + " us ");
        System.out.println("Average time for "+exec/2+" executions (all samples) = "+this.st.getAverage(false)/1000+ " us ");
        System.out.println("Standard desviation for "+exec/2+" executions (all samples) = "+this.st.getDP(false)/1000 + " us ");
        System.out.println("Average time for "+exec+" executions using totalElapsedTime = "+ (totalElapsedTime/exec) / 1000 + " us ");
        System.out.println("Maximum time for "+exec/2+" executions (-10%) = "+this.st.getMax(true)/1000+ " us ");
        System.out.println("Maximum time for "+exec/2+" executions (all samples) = "+this.st.getMax(false)/1000+ " us ");
        System.exit(0);
    }
    
    public void TOMReplyReceive(Object reply){
        long receive_instant = System.nanoTime();        
        try{
            //to compute once a time
            this.mutex.acquire();
        }catch(Exception e){
            e.printStackTrace();
        }
        Object[] obj = (Object[])reply;
        
        if(Integer.parseInt(obj[0].toString()) == currentId){            
            //br.ufsc.das.util.Logger.println("Recebi uma resposta ao pedido com currentId="+currentId); 
            count++;
            //contabiliza a latência quando recebe a f+1-esima resposta
            if(count == f+1){
              //br.ufsc.das.util.Logger.println("Recebi mais do que f respostas ao pedido com currentId="+currentId); 
              if (num_sends>exec/2){
                this.st.store(receive_instant - last_send_instant);
              }
              count = 0;
              currentId+=1;
              this.sm.release();
            }
            //envia o próximo pedido apenas após receber as n respostas
            else if (count == n) {
                
            }
        }
        this.mutex.release();
    }
    
    public static void main(String[] args){
        if (args.length < 2){
            System.out.println("Usage: java TomClientLatency <number of messages> <argument size (bytes)> <interval between requests (ms)>");
            System.exit(1);
        }
        int execs = new Integer(args[0]).intValue();
        int argSize = new Integer(args[1]).intValue();
        int interval = new Integer(args[2]).intValue();
        
        new TOMClientLatency(execs, argSize, interval).run();
    }
    
}
