

/*
 * DeliveryThread.java
 *
 * Created on 18 de Setembro de 2006, 16:47
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package br.ufsc.das.tom.usage;

import br.ufsc.das.paxosatwar.Consensus;
import br.ufsc.das.tom.util.DeliveredEntry;
import br.ufsc.das.tom.util.DeliveredMsgStore;
import br.ufsc.das.tom.util.ProposeValue;
import br.ufsc.das.tom.util.RequestRecoveryMessage;
import br.ufsc.das.tom.util.SetEntry;
import br.ufsc.das.tom.util.TOMMessage;
import br.ufsc.das.util.TOMConfiguration;
import br.ufsc.das.util.*;
import java.util.List;
import java.util.concurrent.LinkedBlockingQueue;

/**
 *
 * @author edualchieri
 */
public class DeliveryThread extends Thread{
    
    private LinkedBlockingQueue<Consensus> decided = new LinkedBlockingQueue<Consensus>();
    private TOMLayer ml;
    private RequestRecover rr;
    private TOMRequestReceiver receiver;
    private DeliveredMsgStore delivered;
    /** Creates a new instance of DeliveryThread */
    public DeliveryThread(TOMLayer ml,TOMRequestReceiver receiver, TOMConfiguration conf) {
        this.ml = ml;
        this.receiver = receiver;
        this.delivered = new DeliveredMsgStore(conf);
        this.rr = new RequestRecover(ml,conf);
    }
    
    public void delivery(Consensus cons){
        try{
            this.decided.put(cons);
            //br.ufsc.das.util.Logger.println("(DeliveryThread) consenso " + cons.getId() + " adicionado ah fila DECIDED ");
        }catch(Exception e){
            e.printStackTrace(System.out);
        }
    }
    
    public void deliverRecoveryRequest(TOMMessage msg){
        if (msg != null){
            this.rr.receive(msg);
            //br.ufsc.das.util.Logger.println("(DeliveryThread) Coloquei na fila um recovery reply ");
        }
        else {
            //br.ufsc.das.util.Logger.println("(DeliveryThread) Recovery reply vem a NULL");
        }
    }
    
    public void computeAck(int consId, int p){
        this.delivered.ack(consId,p);
    }
    
    public SetEntry getRequest(byte[] hash){
        return this.delivered.get(hash);
    }
    
    public void deliverRecoveryRequest(TOMMessage msg, byte[] hash){
        this.rr.receive(msg,hash);
    }
    
    public void checkRequests(List l){
        this.rr.check(l);
    }
    
    public void run(){
        while(true){
            try{
                //br.ufsc.das.util.Logger.println("(DeliveryThread) Antes do DECIDED.take ");
                Consensus cons = decided.take();
                //br.ufsc.das.util.Logger.println("(DeliveryThread) Depois do DECIDED.take apanhou consenso " + cons.getId());
                byte[][] toDeliver = ((ProposeValue)cons.getDecision()).getValues();
                SetEntry[] deliv = new SetEntry[toDeliver.length];
                for(int i = 0; i < toDeliver.length;i++){
                    //br.ufsc.das.util.Logger.println("(DeliveryThread) ("+cons.getId()+") Vou remover hash: ");
                    TOMUtil.print(toDeliver[i]);
                    
                    SetEntry d = ml.unordered.removeHash(toDeliver[i]);
                    //br.ufsc.das.util.Logger.println("(DeliveryThread) ("+cons.getId()+") Hash removido: ");
                    if(d == null){
                        TOMMessage req = rr.checkWithoutWait(
                                                ml.getDN(cons.getId()),toDeliver[i]);
                        if(req != null){
                            d = new SetEntry(req,toDeliver[i]);
                        }else{
                            /** Não envio RR_REQUEST pq sei que vou receber a mensagem do cliente */
                            //br.ufsc.das.util.Logger.println("(DeliveryThread) (i="+ i +") Vou enviar um recovery request");
                            rr.recover(toDeliver[i]);
                           //br.ufsc.das.util.Logger.println("(DeliveryThread) (i="+ i +") Ah espera do recovery reply " );
                             /* */
                            //br.ufsc.das.util.Logger.println("(DeliveryThread) ("+cons.getId()+") Ah espera da mensagem do cliente");
                            req = rr.getMsg();
                            //br.ufsc.das.util.Logger.println("(DeliveryThread) ("+cons.getId()+") (i="+ i +") Recebi mensagem do cliente " );
                            d = new SetEntry(req,toDeliver[i]);
                        }
                        ml.setLastReceived(req.getSender(),req.getSequence());
                    }
                    deliv[i] = d;
                    receiver.TOMRequestReceive(d.msg);
                    ml.delivered(d.msg);
                }
                //br.ufsc.das.util.Logger.println("(DeliveryThread) ("+cons.getId()+") Antes de adicionar ao delivered ");
                if(delivered.add(cons.getId(),deliv)){
                    //br.ufsc.das.util.Logger.println("(DeliveryThread) ("+cons.getId()+") Depois de adicionar ao delivered e antes do removeReady ");
                    DeliveredEntry[] de = delivered.removeReady();
                    //br.ufsc.das.util.Logger.println("(DeliveryThread) ("+cons.getId()+") Depois do removeReady, delivered size= "+de.length);
                    for(int i = 0; i < de.length; i++){
                        ml.getCommunication().send(de[i].getNotAcks(),de[i].getRRMsg());
                    }
                    
                }
                //br.ufsc.das.util.Logger.println("(DeliveryThread) ("+cons.getId()+") Antes do checkTimer ");
                ml.checkTimer();
                ml.lastDecided = cons;
                ml.setLastExec(cons.getId());
                //br.ufsc.das.util.Logger.println("(DeliveryThread) Setting last exec to "+cons.getId());
                ml.removeDN(cons.getId());
                if(cons.getId() > 2){
                    ml.lm.removeStableConsenusInfos(cons.getId() - 3);
                    ml.execManager.setStable(cons.getId() - 3);
                }
                //br.ufsc.das.util.Logger.println("(DeliveryThread) Setting InExec to -1");
                ml.setInExec(-1);
                if(!ml.acceptor.executeAcceptedPendent(cons.getId()+1)){
                    //br.ufsc.das.util.Logger.println("(DeliveryThread) Antes do checkNextExecution");
                    ml.execManager.checkNextExecution(cons.getId()+1);
                }
                //br.ufsc.das.util.Logger.println("(DeliveryThread) All finished");
            }catch(Exception e){
                e.printStackTrace(System.out);
            }
        }
    }
}
