

/*
 * RequestRecover.java
 *
 * Created on 18 de Setembro de 2006, 15:32
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package br.ufsc.das.tom.usage;

import br.ufsc.das.tom.util.RequestRecoveryMessage;
import br.ufsc.das.tom.util.SetEntry;
import static br.ufsc.das.util.TOMUtil.*;
import br.ufsc.das.tom.util.TOMMessage;
import br.ufsc.das.util.TOMConfiguration;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Semaphore;

/**
 *
 * @author edualchieri
 */
public class RequestRecover{
    
    private Semaphore wait = new Semaphore(0);
    private Semaphore mutex = new Semaphore(1);
    private int[] group;
    private TOMMessage msg;
    private TOMLayer ml;
    private byte[] hash;
    private TOMConfiguration conf;
    
    /** Creates a new instance of RequestRecover */
    public RequestRecover(TOMLayer ml, TOMConfiguration conf) {
        this.ml = ml;
        this.conf = conf;
        group = new int[conf.getN()-1];
        int c=0;
        for(int i = 0; i < conf.getN();i++){
            if(i != conf.getProcessId()){
                group[c++] = i;
            }
        }
    }
    
    public void recover(byte[] hash){
        this.msg = null;
        this.hash = hash;
        this.ml.getCommunication().send(group,
                new RequestRecoveryMessage(hash,conf.getProcessId()));
    }
         
        
    public void receive(TOMMessage msg, byte[] msgHash){
        try{
            this.mutex.acquire();
        }catch(Exception e){
            e.printStackTrace();
        }
        if(this.hash != null){
            if(equalsHash(msgHash,hash)){
                this.msg = msg;
                this.hash = null;
                //br.ufsc.das.util.Logger.println("(RequestRecover) Mensagem recebida coincide com a que estavamos ah espera");
                this.wait.release();
            }
            else {
                //br.ufsc.das.util.Logger.println("(RequestRecover) Mensagem recebida NAO coincide com a que estavamos ah espera");
            }
        }
       
        mutex.release();
    }
    
    public TOMMessage checkWithoutWait(List<RequestRecoveryMessage> l, byte[] h){
        try{
            this.mutex.acquire();
        }catch(Exception e){
            e.printStackTrace();
        }
        if(l != null && l.size() > conf.getF()){        
            List<TOMMessage> candidates = new LinkedList<TOMMessage>();
            for(int i = 0; i < l.size(); i++){
                SetEntry[] se = l.get(i).getMsgsDelivered();
                if(se != null){
                    for(int x = 0; x < se.length; x++){
                        if(equalsHash(se[x].hash,h)){
                            candidates.add(se[x].msg);
                            break;
                        }
                    }
                }
            }
            TOMMessage result = getFromCandidates(candidates);
            if(result != null){
                this.mutex.release();
                return result;
            }
        }
        else if (l!=null){
            //br.ufsc.das.util.Logger.println("(RequestRecover) checkWithoutWait: l.size="+l.size());
        }
        this.mutex.release();
        return null;
    }
    
    
    public void check(List<RequestRecoveryMessage> l){
        try{
            this.mutex.acquire();
        }catch(Exception e){
            e.printStackTrace();
        }
        if(this.hash != null && l != null && l.size() > conf.getF()){
            List<TOMMessage> candidates = new LinkedList<TOMMessage>();
            for(int i = 0; i < l.size(); i++){
                SetEntry[] se = l.get(i).getMsgsDelivered();
                if(se != null){
                    for(int x = 0; x < se.length; x++){
                        if(equalsHash(se[x].hash,this.hash)){
                            candidates.add(se[x].msg);
                            break;
                        }
                    }
                }
            }
            TOMMessage result = getFromCandidates(candidates);
            if(result != null){
                this.msg = result;
                this.hash = null;
                this.wait.release();
            }
        }
        this.mutex.release();
    }
    
    private TOMMessage getFromCandidates(List<TOMMessage> candidates){
        TOMMessage comp = null;
        for(int i = 0; i < candidates.size();i++){
            comp = candidates.get(i);
            int c = 1;
            for(int j = i+1; j < candidates.size();j++){
                if(comp.equals(candidates.get(j))){
                    c++;
                }
            }
            if(c > conf.getF()){
                return comp;
            }
        }
        return null;
    }
    
    
    public void receive(TOMMessage msg){
        try{
            this.mutex.acquire();
        }catch(Exception e){
            e.printStackTrace();
        }
        if(this.hash != null){
            byte[] recHash = ml.getHash(msg);
            if(equalsHash(recHash,hash)){
                this.msg = msg;
                this.hash = null;
                this.wait.release();
            }
        }
        else
            //br.ufsc.das.util.Logger.println("(RequestRecover) hash do RR_REPLY não é igual ao do RR_REQUEST");
        mutex.release();
    }
    
    public TOMMessage getMsg(){        
        try{
            this.wait.acquire();
        }catch(Exception e){
            e.printStackTrace();
        }
        
        return this.msg;
    }
    
}
