

package br.ufsc.das.tom.usage;
import br.ufsc.das.paxosatwar.Consensus;
import br.ufsc.das.paxosatwar.roles.Acceptor;
import br.ufsc.das.paxosatwar.roles.util.Execution;
import br.ufsc.das.paxosatwar.roles.util.ExecutionManager;
import br.ufsc.das.paxosatwar.roles.util.LeaderModule;
import br.ufsc.das.paxosatwar.roles.util.Round;
import br.ufsc.das.tom.timer.RTInfo;
import br.ufsc.das.tom.timer.RTTimer;
import br.ufsc.das.tom.util.ProposeValue;
import br.ufsc.das.tom.timer.RTLeaderChange;
import br.ufsc.das.tom.util.RequestRecoveryMessage;
import br.ufsc.das.tom.timer.RTCollect;
import br.ufsc.das.tom.util.SetEntry;
import br.ufsc.das.tom.util.TOMMessage;
import br.ufsc.das.tom.util.TOMSet;
import br.ufsc.das.tom.timer.RTMessage;
import br.ufsc.das.util.Configuration;
import br.ufsc.das.util.CommunicationSystem;
import br.ufsc.das.util.TOMConfiguration;
import br.ufsc.das.util.TOMUtil;
import java.io.ByteArrayOutputStream;
import java.io.ObjectOutputStream;
import java.security.MessageDigest;
import java.security.SignedObject;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.Timer;
import java.util.TreeSet;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.ReentrantLock;
import static br.ufsc.das.util.TOMUtil.*;

public class TOMLayer extends Thread{
    
    private static int maxPropSize = 100;
    
    private long deliveryInterval;
    public TOMSet unordered = new TOMSet();
    private Map<Integer,Integer> lastReceived = new HashMap<Integer,Integer>();
    private Map<Integer,List> deliveryNotified = new HashMap<Integer,List>();
    private Map<Integer,List> timeoutInfo = new HashMap<Integer,List>();
    public ExecutionManager execManager;
    private int inExecution = -1;
    private int lastExecuted = -1;
    public LeaderModule lm;
    public Consensus lastDecided; // not stable (only this)
    public Acceptor acceptor;
    public RTTimer timer;
    private CommunicationSystem communication;
    private DeliveryThread dt;
    private MessageDigest md;
    private ByteArrayOutputStream bOut;
    private ObjectOutputStream obOut;
    private ReentrantLock lockLR;
    private ReentrantLock lockDN;
    private ReentrantLock lockTI;
    
    private TOMConfiguration conf;
    
    /**
     * Creates a new instance of TOMulticastLayer
     */
    public TOMLayer(ExecutionManager manager, TOMRequestReceiver receiver,
            LeaderModule lm, Acceptor a, CommunicationSystem cs, TOMConfiguration conf) {
        this.execManager = manager;
        this.deliveryInterval = conf.getTOMPeriod();
        this.lm = lm;
        this.acceptor = a;
        this.communication = cs;
        this.conf = conf;
        this.timer = new RTTimer(this);
        this.lockLR = new ReentrantLock();
        this.lockDN = new ReentrantLock();
        this.lockTI = new ReentrantLock(); 
        this.maxPropSize = conf.getMaxBatchSize();
        try{
            this.md = MessageDigest.getInstance("MD5");
            this.bOut = new ByteArrayOutputStream();
            this.obOut = new ObjectOutputStream(bOut);
        }catch(Exception e){
            e.printStackTrace(System.out);
        }
        this.dt = new DeliveryThread(this,receiver,conf);
        this.dt.start();
    }
    
    public TOMConfiguration getConf(){
        return this.conf;
    }
    
    public byte[] getHash(TOMMessage value){
        byte[] ret = null;
        try{
            obOut.writeObject(value);
            ret = md.digest(bOut.toByteArray());
            md.reset();
            obOut.reset();
            bOut.reset();
        }catch(Exception e){
            e.printStackTrace(System.out);
        }
        return ret;
    }
    
    public CommunicationSystem getCommunication(){
        return this.communication;
    }
    
    public void setLastExec(int last){
        this.lastExecuted = last;
    }
    
    public int getLastExec(){
        return this.lastExecuted;
    }
    
    public void setInExec(int inEx){
        this.inExecution = inEx;
    }
    
    public int getInExec(){
        return this.inExecution;
    }
    
    private ProposeValue createPropose(){
        int tam = 0;
        if(unordered.size() <= maxPropSize){
            tam = unordered.size();
        }else{
            tam = maxPropSize;
        }
        byte[][] prop = new byte[tam][];
        for(int i = 0; i < tam;i++){
            prop[i] =  unordered.getHash(i);
        }
        return new ProposeValue(prop);
    }
    
    public void run() {
        while(true) {
            ////br.ufsc.das.util.Logger.println("(TOMLayer) unordered size="+unordered.size());
            if(unordered.size() < conf.getMaxBatchSize()){
                synchronized(this) {
                    try{
                        ////br.ufsc.das.util.Logger.println("(TOMLayer) Waiting " + deliveryInterval + " ms");
                        this.wait(deliveryInterval);
                    }catch(InterruptedException ie) {
                        ie.printStackTrace(System.out);
                    }
                }
            }
            if((getInExec() == -1)
            && (lm.getLeader(getLastExec()+1,0) == conf.getProcessId())
            && (!unordered.isEmpty())){
                int execId = getLastExec()+1;
                setInExec(execId);
                ////br.ufsc.das.util.Logger.println("(TOMLayer) Ah espera do mutex antes do getExecution");
                acceptor.getMEZone();
                Execution exec = execManager.getExecution(execId);
                acceptor.leaveMEZone();
                ////br.ufsc.das.util.Logger.println("(TOMLayer) Depois do getExecution, vou criar o propose");
                exec.getLearner().propose(createPropose());
            }
            else if (lm.getLeader(getLastExec()+1,0) == conf.getProcessId() && !unordered.isEmpty()){
                //br.ufsc.das.util.Logger.println("(TOMLayer) Sou o lider, o unordered nao esta vazio e nao entrei no if da execucao pq ");
                //br.ufsc.das.util.Logger.println(" getInExec()=" + getInExec());              
            }
        }
    }
    
    //called by the delivery thread to infor that a msg was delivered
    public void delivered(TOMMessage msg){
        int reqId = msg.getSender()+msg.getSequence();
        timer.cancelTimer(reqId);
        removeTimeoutInfo(reqId);
    }
    
    public void decided(Consensus cons){
        this.dt.delivery(cons);
    }
    
    public void computeAck(int consId, int p){
        this.dt.computeAck(consId,p);
    }
    
    public void checkTimer(){
        SetEntry se = null;
        for(int i = 0; i < unordered.size(); i++){
            se = (SetEntry)unordered.get(i);
            if(timer.monitoring(se.msg.getSender()+se.msg.getSequence())){
                break;
            }
        }
    }
    
    private void checkPendentExecution(){
        //br.ufsc.das.util.Logger.println("(TOMLayer) checkPendentExecution: ah espera do mutex");
        acceptor.getMEZone();
        //br.ufsc.das.util.Logger.println("(TOMLayer) checkPendentExecution: consegui o mutex");
        if(getInExec() != -1){
            Execution exec = execManager.getExecution(getInExec());
            Round r = exec.getRound(0); // s� vai acontecer com o round 0;
            if(containRequests(r) && r.propValue!=null){
                //br.ufsc.das.util.Logger.println("(TOMLayer) checkPendentExecution: tenho pedidos pendentes e o valor nao eh null, logo vou entrar no weakAccept");
                acceptor.weakAccept(r);
            }
        }
        acceptor.leaveMEZone();
        //br.ufsc.das.util.Logger.println("(TOMLayer) checkPendentExecution: libertei o mutex");
    }
    
    public boolean containRequests(Round r){
        ProposeValue prop = ((ProposeValue)r.propValue);
        if(prop != null){
            byte[][] values = prop.getValues();
            for(int i = 0; i < values.length; i++){
                if(!unordered.containsHash(values[i])){
                    return false;
                }
            }
        }
        return true;
    }
    
    public boolean containRequest(byte[] hash){
        return unordered.containsHash(hash);
    }
    
    public void discartRequest(byte[] msgHash){
        SetEntry se = unordered.removeHash(msgHash);
        if(se != null && se.msg != null){
            int reqId = se.msg.getSender()+se.msg.getSequence();
            removeTimeoutInfo(reqId);
            timer.cancelTimer(reqId);
        }
    }
    
    public void discartRequest(TOMMessage msg){
        //descartar tamb�m possiveis dados referentes a timeouts
        unordered.removeMsg(msg);
        int reqId = msg.getSender()+msg.getSequence();
        removeTimeoutInfo(reqId);
        timer.cancelTimer(reqId);
    }
    
    public void discartRequest(int reqId){
        //descartar tamb�m possiveis dados referentes a timeouts
        unordered.removeMsg(reqId);
        removeTimeoutInfo(reqId);
        timer.cancelTimer(reqId);
    }
    
    public boolean requestTimeout(int reqId, int timeout){
        //verifica se ja n�o foi ordenada
        if(unordered.containsMsg(reqId)){
            if(timeout > conf.getF()){
                discartRequest(reqId);
                return false;
            }else{
                RTInfo rti = getTimeoutInfo(reqId,timeout);
                if(!rti.getTimeout(conf.getProcessId())){
                    rti.timeout(conf.getProcessId());
                    ////br.ufsc.das.util.Logger.println("Enviando msg de timeout: reqId: "+reqId+" timeout: "+timeout);
                    sendTimeoutMessage(reqId,timeout);
                    return true;
                }
            }
        }
        return false;
    }
    
    public void sendTimeoutMessage(int reqId, int timeout){
        RTMessage rtm =
                new RTMessage(RT_TIMEOUT,reqId,timeout,conf.getProcessId(),null);
        this.communication.send(execManager.getOtherAcceptors(),rtm);
    }
    
    public void sendCollectMessage(int reqId, int timeout, RTCollect collect){
        RTMessage rtm =
                new RTMessage(RT_COLLECT,reqId,timeout,
                conf.getProcessId(),acceptor.sign(collect));
        if(collect.getNewLeader() == conf.getProcessId()){
            getTimeoutInfo(reqId,timeout).collect((SignedObject)rtm.getContent(),
                    conf.getProcessId());
        }else{
            int[] target = {collect.getNewLeader()};
            this.communication.send(target,rtm);
        }
    }
    
    public void sendNewLeaderMessage(int reqId, int timeout, RTLeaderChange rtLC){
        RTMessage rtm =
                new RTMessage(RT_LEADER,reqId,timeout,
                conf.getProcessId(),rtLC);
        //br.ufsc.das.util.Logger.println("Atualizando leader para "+rtLC.newLeader+" a partir de "+rtLC.start);
        updateLeader(reqId,rtLC.start,rtLC.newLeader,timeout);
        
        this.communication.send(execManager.getOtherAcceptors(),rtm);
    }
    
    private void updateLeader(int reqId, int start, int newLeader, int timeout){
        lm.addLeaderInfo(start,0,newLeader);
        this.removeTimeoutInfo(reqId,timeout);
        timer.restartMonitoring(reqId);
        execManager.restar();
    }
    
    public void deliverTimeoutRequest(RTMessage msg){
        try{
            if(msg.getRTType() == RT_TIMEOUT){
                ////br.ufsc.das.util.Logger.println(msg.getFrom()+ " 1 recebendo msg de timeout: reqId: "+msg.getReqId()+" timeout: "+msg.getTimeout());
                if(unordered.containsMsg(msg.getReqId())){
                    ////br.ufsc.das.util.Logger.println("2 recebendo msg de timeout: reqId: "+msg.getReqId()+" timeout: "+msg.getTimeout());
                    getTimeoutInfo(msg.getReqId(),msg.getTimeout()).timeout(msg.getFrom());
                }
            }else if(msg.getRTType() == RT_COLLECT){
               ////br.ufsc.das.util.Logger.println(msg.getReqId()+" 1 recebendo collect de "+msg.getFrom());
                SignedObject so = (SignedObject)msg.getContent();
                if(this.acceptor.verifySignature(so,msg.getFrom())){
                    //br.ufsc.das.util.Logger.println("2 recebendo collect de "+msg.getFrom());
                    RTCollect rtc = (RTCollect)so.getObject();
                    int reqId = rtc.getReqId();
                    int nl = (reqId+rtc.getTimeout()-1)%conf.getN();
                    if(nl == conf.getProcessId() && nl == rtc.getNewLeader()){
                        ////br.ufsc.das.util.Logger.println("3 recebendo collect de "+msg.getFrom());
                        getTimeoutInfo(reqId,rtc.getTimeout()).collect(so,msg.getFrom());
                    }
                }
            }else if(msg.getRTType() == RT_LEADER){
                //br.ufsc.das.util.Logger.println("1 recebendo newLeader de "+msg.getFrom());
                RTLeaderChange rtLC = (RTLeaderChange)msg.getContent();
                RTCollect[] rtc = getValid(msg.getReqId(),msg.getTimeout(),rtLC.proof);
                if(goodStartLeader(rtc,conf.getF(),rtLC.start,rtLC.newLeader)){
                   //br.ufsc.das.util.Logger.println("Atualizando leader para "+rtLC.newLeader+" a partir de "+rtLC.start);
                    updateLeader(msg.getReqId(),rtLC.start,rtLC.newLeader,msg.getTimeout());
                    //FALTA... eliminar dados referentes a consensos maiores q start.
                }
            }
        }catch(Exception e){
            e.printStackTrace(System.out);
        }
    }
    
    
    private RTCollect[] getValid(int reqId, int timeout, SignedObject[] proof){
        Collection<RTCollect> valid = new HashSet<RTCollect>();
        try{
            for(int i = 0; i < proof.length; i++){
                if(proof[i] != null && acceptor.verifySignature(proof[i],i)){
                    RTCollect rtc = (RTCollect)proof[i].getObject();
                    if(rtc != null && rtc.getReqId() == reqId && rtc.getTimeout() == timeout){
                        valid.add(rtc);
                    }
                }
            }
        }catch(Exception e){
            e.printStackTrace(System.out);
        }
        return valid.toArray(new RTCollect[0]);
    }
    
    public void deliver(TOMMessage msg) {
        if(msg == null){ //s� para proteger
            //br.ufsc.das.util.Logger.println("(TOMLayer)TOMMessage vem a null");
            return;
        }
        /*if(conf.getProcessId() == 0 && msg.getSequence() == 3){
            return;
        }*/
        lockLR.lock();
        if(!(lastReceived.containsKey(msg.getSender()))
        || (lastReceived.get(msg.getSender()) + 1 == msg.getSequence())){
            
            lastReceived.put(msg.getSender(),msg.getSequence());
            lockLR.unlock();
            byte[] hash = getHash(msg);
            //br.ufsc.das.util.Logger.println("(TOMLayer) A guardar hash da msg "+msg.getSequence()+ " enviada por " + msg.getSender() + ", hash=");
            TOMUtil.print(hash);
            //br.ufsc.das.util.Logger.println("(TOMLayer) ("+msg.getSequence()+ ") A adicionar ao unordered");
            unordered.addElement(new SetEntry(msg,hash));
            //br.ufsc.das.util.Logger.println("(TOMLayer) ("+msg.getSequence()+ ") A verificar se existe algum RR ah espera");
            this.dt.deliverRecoveryRequest(msg,hash);
            //br.ufsc.das.util.Logger.println("(TOMLayer) ("+msg.getSequence()+ ") A verificar execucoes pendentes");
            checkPendentExecution();
            //br.ufsc.das.util.Logger.println("(TOMLayer) ("+msg.getSequence()+ ") A verificar o timer");
            checkTimer();
        }
        else {
            //br.ufsc.das.util.Logger.println("(TOMLayer) TOMMessage recebida nao eh a seguinte da ultima recebida, recebi "+msg.getSequence()+" mas estava ah espera de receber a "+(lastReceived.get(msg.getSender()) + 1));
            lockLR.unlock();
        }
    }
    
    public void setLastReceived(int sender, int seq){
        lockLR.lock();
        if(!(lastReceived.containsKey(sender))
        || (lastReceived.get(sender) + 1 == seq)){
            lastReceived.put(sender,seq);
        }
        lockLR.unlock();
    }
    
    
    public void deliverRecoveryRequest(RequestRecoveryMessage rrm){
        if(rrm.getFrom() > -1 && rrm.getFrom() < conf.getN()){
            if(rrm.getId() ==  RR_REQUEST){//request
                //br.ufsc.das.util.Logger.println("(TOMLayer) A processar RR_REQUEST enviado pelo " + rrm.getFrom());
                SetEntry se = dt.getRequest(rrm.getHash());
                if(se == null){
                    se = unordered.getHash(rrm.getHash());
                }
                if(se != null){
                    //br.ufsc.das.util.Logger.println("(TOMLayer) A enviar RR_REPLY para " + rrm.getFrom());
                    int[] target = {rrm.getFrom()};
                    this.communication.send(target,
                            new RequestRecoveryMessage(se.msg,conf.getProcessId()));
                }
                else {
                    //br.ufsc.das.util.Logger.println("(TOMLayer) Não envio RR_REPLY pq não tenho a mensagem");
                }
            }else if(rrm.getId() ==  RR_REPLY){//reply
                //br.ufsc.das.util.Logger.println("(TOMLayer) A processar RR_REPLY enviado por "+rrm.getFrom());
                this.dt.deliverRecoveryRequest(rrm.getMsg());
            }else if(rrm.getId() ==  RR_DELIVERED){//delivered
                //br.ufsc.das.util.Logger.println("(TOMLayer) A processar RR_DELIVERED enviado por "+rrm.getFrom());
                if(rrm.getConsId() > getLastExec() &&
                        rrm.getConsId() <= (10*conf.getPaxosHighMarc()+getLastExec())){
                    SetEntry[] msgs = rrm.getMsgsDelivered();
                    List<RequestRecoveryMessage> l = getDN(rrm.getConsId());
                    boolean add = true;
                    for(int i = 0; i < l.size(); i++){
                        if(l.get(i).getFrom() == rrm.getFrom()){
                            add = false;
                            break;
                        }
                    }
                    if(add){
                        l.add(rrm);
                        if(rrm.getConsId() == getInExec()){
                            this.dt.checkRequests(l);
                        }
                    }
                }
            }
        }
    }
    
    public List getDN(int id){
        lockDN.lock();
        List l = deliveryNotified.get(id);
        if(l == null){
            l = new LinkedList();
            deliveryNotified.put(id,l);
        }
        lockDN.unlock();
        return l;
    }
    
    public RTInfo getTimeoutInfo(int reqId, int timeoutNum){
        List<RTInfo> l = getTimeoutInfo(reqId);
        for(int i = 0; i < l.size(); i++){
            if(l.get(i).getTimeoutNum() == timeoutNum){
                return l.get(i);
            }
        }
        RTInfo ret = new RTInfo(this.conf,reqId,timeoutNum,this);
        l.add(ret);
        return ret;
    }
    
    private void removeTimeoutInfo(int reqId){
        lockTI.lock();
        this.timeoutInfo.remove(reqId);
        lockTI.unlock();
    }
    
    
    private void removeTimeoutInfo(int reqId, int timeoutNum){
        List<RTInfo> l = getTimeoutInfo(reqId);
        for(int i = 0; i < l.size(); i++){
            if(l.get(i).getTimeoutNum() == timeoutNum){
                l.remove(i);
            }
        }
    }
    
    
    public List getTimeoutInfo(int reqId){
        lockTI.lock();
        List l = timeoutInfo.get(reqId);
        if(l == null){
            l = new LinkedList();
            timeoutInfo.put(reqId,l);
        }
        lockTI.unlock();
        return l;
    }
    
    
    public void removeDN(int id){
        lockDN.lock();
        deliveryNotified.remove(id);
        lockDN.unlock();
    }
}
