

/*
 * DeliveredMsgStore.java
 *
 * Created on 25 de Setembro de 2006, 15:20
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package br.ufsc.das.tom.util;

import br.ufsc.das.util.TOMConfiguration;
import br.ufsc.das.util.TOMUtil;

import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.locks.ReentrantLock;

/**
 *
 * @author edualchieri
 */
public class DeliveredMsgStore {
    
    private List<DeliveredEntry> delivered = new LinkedList<DeliveredEntry>();
    private TOMConfiguration conf;
    
    private ReentrantLock lock;
    
    /** Creates a new instance of DeliveredMsgStore */
    public DeliveredMsgStore(TOMConfiguration conf) {
        this.conf = conf;    
        lock = new ReentrantLock();
    }
    
    public boolean add(int consId, SetEntry[] msgs){
        lock.lock();
        delivered.add(new DeliveredEntry(consId,msgs,conf));
        if(delivered.size() > conf.getPaxosHighMarc()){
            int lim = conf.getN() - conf.getF();
            for(int i = 0; i < delivered.size(); i++){
                if(delivered.get(i).countAcks() >= lim){
                    lock.unlock();
                    return true;
                }
            }
        }
        lock.unlock();
        return false;
    }
  
    public void ack(int consId, int processId){
        lock.lock();
        DeliveredEntry e = getEntry(consId);
        if(e != null && e.ack(processId)){
            delivered.remove(e);
        }
        lock.unlock();
    }
    
    private DeliveredEntry getEntry(int consId){
       
        DeliveredEntry e = null;
        for(int i = 0; i < delivered.size(); i++){
            e = delivered.get(i);
            if(e != null && e.consId == consId){
                return e;
            }
        }
       
        return null;
    }
    
    public DeliveredEntry[] removeReady(){
        List<DeliveredEntry> ret = new LinkedList();
        int lim = conf.getN() - conf.getF();
        lock.lock();
        for(int i = 0; i < delivered.size(); i++){
            if(delivered.get(i).countAcks() >= lim){
                ret.add(delivered.get(i));
            }
        }
        delivered.removeAll(ret);
        lock.unlock();
        return ret.toArray(new DeliveredEntry[0]);
    }
    
    public SetEntry get(byte[] msgHash){        
        lock.lock();
        for(int i = 0; i < delivered.size(); i++){
            DeliveredEntry e = delivered.get(i);            
            for(int j = 0; j < e.msgs.length; j++){
                if(TOMUtil.equalsHash(msgHash,e.msgs[j].hash)){                                        
                    lock.unlock();
                    return e.msgs[j];
                }
            }
        }
        lock.unlock();
        return null;
    }
}
