

/*
 * RecoveryRequestMessage.java
 *
 * Created on 18 de Setembro de 2006, 15:14
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package br.ufsc.das.tom.util;

import br.ufsc.das.tom.usage.TOMLayer;
import br.ufsc.das.util.Configuration;
import br.ufsc.das.util.CommunicationSystem;
import br.ufsc.das.util.SystemMessage;
import static br.ufsc.das.util.TOMUtil.*;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

/**
 *
 * @author edualchieri
 */
public class RequestRecoveryMessage extends SystemMessage{
    
    private int id;
    private int consId;
    private byte[] hash;
    private TOMMessage msg;
    private int from;
    private SetEntry[] delivered;
    
    public RequestRecoveryMessage(){}
    
    /** Creates a new instance of RecoveryRequestMessage */
    public RequestRecoveryMessage(byte[] hash, int from) {
        super(CommunicationSystem.RR_MSG);
        this.hash = hash;
        this.id = RR_REQUEST;
        this.from = from;
    }
    
    public RequestRecoveryMessage(TOMMessage msg, int from) {
        super(CommunicationSystem.RR_MSG);
        this.msg = msg;
        this.id = RR_REPLY;
        this.from = from;
    }
    
    public RequestRecoveryMessage(SetEntry[] delivered, int from, int consId) {
        super(CommunicationSystem.RR_MSG);
        this.delivered = delivered;
        this.id = RR_DELIVERED;
        this.from = from;
        this.consId = consId;
    }
    
    
    public void writeExternal(ObjectOutput out)
    throws IOException{
        super.writeExternal(out);
        out.writeInt(consId);
        out.writeInt(id);
        if(hash != null){
            out.writeInt(hash.length);
            out.write(hash);
        }else{
            out.writeInt(-1);
        }
        out.writeObject(msg);
        out.writeInt(from);
        out.writeObject(delivered);
    }
    
    public void readExternal(ObjectInput in)
    throws IOException,
            ClassNotFoundException{
        super.readExternal(in);
        consId = in.readInt();
        id = in.readInt();
        int t =in.readInt();
        if(t>0){
            hash = new byte[t];
            in.readFully(hash);
        }else{
            hash = null;
        }
        msg = (TOMMessage)in.readObject();
        from = in.readInt();
        delivered = (SetEntry[]) in.readObject();
    }
    
    public int getConsId(){
        return this.consId;
    }
    
    public int getId(){
        return this.id;
    }
    
    public int getFrom(){
        return this.from;
    }
    
    public TOMMessage getMsg(){
        return this.msg;
    }
    
    public SetEntry[] getMsgsDelivered(){
        return this.delivered;
    }
    
    public byte[] getHash(){
        return this.hash;
    }
    
    public String toString() {
        return "consId="+getConsId()+", type="+getId()+", from="+getFrom();
    }
    
}
