package br.ufsc.das.tom.util;

import br.ufsc.das.util.CommunicationSystem;
import br.ufsc.das.util.SystemMessage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectInputStream;
import java.io.ObjectOutput;
import java.io.ObjectOutputStream;


public class TOMMessage extends SystemMessage {
    
    private int sender;
    private int sequence;
    private Object content;
    
    public TOMMessage(){}
    
    /** Creates a new instance of TOMulticastContent */
    public TOMMessage(int sender, int sequence, Object content, int type) {
        super(type);
        this.sender = sender;
        this.sequence = sequence;
        this.content = content;
    }
    
    
    public void writeExternal(ObjectOutput out)
    throws IOException{
        super.writeExternal(out);
        out.writeInt(sender);
        out.writeInt(sequence);
        out.writeObject(content);
        
    }
    
    public void readExternal(ObjectInput in)
    throws IOException,
            ClassNotFoundException{
        super.readExternal(in);
        sender = in.readInt();
        sequence = in.readInt();
        content = in.readObject();
    }
    
    public Object clone(){
        return null;
    }
    
    public int getSender() {
        return sender;
    }
    
    public int getSequence() {
        return sequence;
    }
    
    public Object getContent() {
        return content;
    }
    
    /**
     * Verifies if two objects are equal.
     *
     * Two TOMMessage are equal if they have the same sender,
     * sequence number and content.
     */
    public boolean equals(Object o) {
        if(o == null) {
            return false;
        }
        
        if(!(o instanceof TOMMessage)) {
            return false;
        }
        
        TOMMessage mc = (TOMMessage)o;
        
        return (mc.getSender() == sender) && (mc.getSequence() == sequence) &&
                 mc.getContent().equals(this.content);
    }
    
    public String toString() {
        return sender+":"+sequence;
    }
}
