/*
 * TOMConfiguration.java
 *
 * Created on 27 de Setembro de 2006, 18:03
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package br.ufsc.das.util;

import java.security.PrivateKey;
import java.security.PublicKey;

/**
 *
 * @author edualchieri
 */
public class TOMConfiguration extends Configuration{
    
    protected int n;
    protected int f;
    protected int requestTimeout;
    protected int freezeInitialTimeout;
    protected int tomPeriod;
    protected int paxosHighMarc;
    protected static RSAKeysLoader rsaLoader;
    protected Logger log;
    protected int maxBatchSize;
    
    
    /** Creates a new instance of TOMConfiguration */
    public TOMConfiguration(int processId) {
        super(processId);        
    }
    
     /** Creates a new instance of TOMConfiguration */
    public TOMConfiguration(int processId, String configHome) {
        super(processId,configHome);
    }
    
    protected void init(){
        super.init();
        try{
            n = Integer.parseInt(configs.remove("system.servers.num").toString());
            
            String s = (String) configs.remove("system.servers.f");
            if(s == null){
                f = (int)Math.ceil((n-1)/3);    
            }else{
                f = Integer.parseInt(s);
            }
            s = (String) configs.remove("system.paxos.freeze.timeout");
            if(s == null){
                freezeInitialTimeout = n*10;
            }else{
                freezeInitialTimeout = Integer.parseInt(s);;
            }
            s = (String) configs.remove("system.request.timeout");
            if(s == null){
                requestTimeout = freezeInitialTimeout*10;
            }else{
                requestTimeout = Integer.parseInt(s);
            }
            
            s = (String) configs.remove("system.totalordermulticast.period");
            if(s == null){
                tomPeriod = n*5;
            }else{
                tomPeriod = Integer.parseInt(s);;
            }
            s = (String) configs.remove("system.paxos.highMarc");
            if(s == null){
                paxosHighMarc = 10;
            }else{
                paxosHighMarc = Integer.parseInt(s);
                if(paxosHighMarc < 10){
                    paxosHighMarc = 10;
                }
            }
            
            s = (String) configs.remove("system.totalordermulticast.maxbatchsize");
            if(s == null){
                maxBatchSize = 100;
            }else{
                maxBatchSize = Integer.parseInt(s);
            }
            
            rsaLoader = new RSAKeysLoader(this,this.configHome);
        }catch(Exception e){
            System.err.println("Wrong system.config file format.");
            e.printStackTrace();
        }
    }
    
     public int getRequestTimeout(){
        return requestTimeout;
    }
    
    public int getN(){
        return n;
    }
    
    public int getF(){
        return f;
    }
    
    public int getTOMPeriod(){
        return tomPeriod;
    }
    
    public int getFreezeInitialTimeout(){
        return freezeInitialTimeout;
    }
    
    public int getPaxosHighMarc(){
        return paxosHighMarc;
    }
    
    public int getMaxBatchSize(){
        return maxBatchSize;
    }
    
    
    public PublicKey[] getRSAPublicKeys(){
        try{
            return rsaLoader.loadPublicKeys();
        }catch(Exception e){
            e.printStackTrace();
            return null;
        }
    }
    
    public static PrivateKey getRSAPrivateKey(){
        try{
            return rsaLoader.loadPrivateKey();
        }catch(Exception e){
            e.printStackTrace();
            return null;
        }
    }
    
}
